#include <stdio.h>

#include <string.h>
#include <assert.h>

#include "Bit.h"
void
initStates (struct character *c, int nums)
{				/* character has num states */
  int i;
  c->nums = NUMSLOTS (nums);
  c->bits = nums;
  c->states = (char *) malloc (sizeof (char) * c->nums);
  for (i = 0; i < c->nums; i++)
    {
      c->states[i] = '\0';
    }
  return;
}

void
clearCharAll (struct character *c)
{				/* clear all bits to zero */
  int i;
  for (i = 0; i < c->nums; i++)
    c->states[i] = '\0';
}

void
charAnd (struct character *c1, struct character *c2, struct character *c3)
{
/*c1: the result of AND,
 *  * c2/c3, the two characters to perform this operation*/
  int i;
  assert (c3->nums == c2->nums && c2->nums == c1->nums);
  for (i = 0; i < c1->nums; i++)
    {
      c1->states[i] = c2->states[i] & c3->states[i];
    }
  return;
}

void
charOR (struct character *c1, struct character *c2, struct character *c3)
{
/*c1: the result of OR,
 *  *  * c2/c3, the two characters to perform this operation*/
  int i;
  assert (c3->nums == c2->nums && c2->nums == c1->nums);
  for (i = 0; i < c1->nums; i++)
    {
      c1->states[i] = c2->states[i] | c3->states[i];
    }
  return;
}

int
allZero (struct character *c1)
{
/* test if c1 is all zero, i.e., empty*/
  int result = 1, i;
  for (i = 0; i < c1->nums; i++)
    {
      if (c1->states[i] != '\0')
	return 0;
    }
  return result;		/* if all zero, return 1 */
}

int
randomPick (struct character *c1)
{
/* randomly pick a '1' location in c1 and set all others to zero*/

  /*algorithm: jump to a location of the array, and move either 
   * left or right to search the first one that is not 1,
   * hope this will be a fairly random procedure*/
  int loc, i, j;

  loc = random () % c1->bits;
  if (random () % 2 == 0)
    {				/* move to left and find the closest one bit
				   that is not 0 */
      while (loc >= 0)
	{
	  if (TEST (c1->states, loc))
	    {
	      return loc;
	    }
	  loc--;
	}
      /* if not find 1s, start to the right */
      while (loc < c1->bits)
	{
	  if (TEST (c1->states, loc))
	    {
	      return loc;
	    }
	  loc++;
	}
    }
  else
    {				/* move to the right */
      while (loc < c1->bits)
	{
	  if (TEST (c1->states, loc))
	    {
	      return loc;
	    }
	  loc++;
	}
      /* if not find, start to the left */
      while (loc >= 0)
	{
	  if (TEST (c1->states, loc))
	    {
	      return loc;
	    }
	  loc--;
	}
    }
  printf ("Something wrong, you should not reach here\n");
  printf ("At least one bit should be 1\n");
  exit (-1);
}

void
setTrue (struct character *c1, int loc)
{
  if (loc >= c1->bits)
    return;
  SET (c1->states, loc);
}

/*test if loc is 1*/
int
testTrue (struct character *c1, int loc)
{
  if (loc >= c1->bits)
    return 0;
  return TEST (c1->states, loc);
}

void
copyState (struct character *c1, struct character *c2)
{
  memcpy (c2->states, c1->states, c2->nums);
}

#if 0
main ()
{
  int MAX = 100;
  char arr[NUMSLOTS (MAX)];
  int i, j;
  memset (arr, 0, NUMSLOTS (MAX));
/*    for ( i = 2 ; i < MAX ; i++ )
    {
        if ( !TEST ( arr, i ) )
        {
            printf ( "\n%d", i ) ;
            for ( j = i + i ; j < MAX ; j += i )
                SET ( arr, j ) ;
        }
    }*/
  SET (arr, 0);

}
#endif
