#ifndef BIT_H_
#define BIT_H_

#define CHARSIZE 8

#define MASK(y) ( 1 << y % CHARSIZE )

#define BITSLOT(y) ( y / CHARSIZE)

#define SET( x, y) ( x[BITSLOT(y)] |= MASK(y) )

#define CLEAR(x, y) ( x[BITSLOT(y)] &= ~MASK(y) )

#define TEST(x, y) ( x[BITSLOT(y)] & MASK(y) )

#define NUMSLOTS(n) ((n + CHARSIZE - 1) / CHARSIZE)
/*
 char arr[NUMSLOTS(50)];
 memset ( arr, 0, NUMSLOTS(50) ) ;
 SET ( arr, 20 ) ; set 21st bit as 1, 
 if ( TEST ( arr, 40 ) ); check if 41st is 1
 */
struct character
{
  char *states;
  unsigned short nums;		/* num of chars */
  unsigned short bits;		/* num of bits */
};

void initStates (struct character *, int nums);
    /* character has num states */

void clearCharAll (struct character *c);

void charAnd (struct character *c1,
	      struct character *c2, struct character *c3);
/*c1: the result of AND, 
 * c2/c3, the two characters to perform this operation*/

void charOR (struct character *c1,
	     struct character *c2, struct character *c3);
/*c1: the result of OR,
 *  * c2/c3, the two characters to perform this operation*/

int allZero (struct character *c1);
/* test if c1 is all zero, i.e., empty*/

int randomPick (struct character *c1);
/* randomly pick a '1' location in c1 and set all others to zero*/

void setTrue (struct character *c1, int loc);

/*test if loc is 1*/
int testTrue (struct character *c1, int loc);

void copyState (struct character *c1, struct character *c2);
/*copy states from one c to another c*/
#endif
