#include "MP.h"
#include "MPTsp.h"
#include <stdlib.h>
#include "Bit.h"
#include "structs.h"
#include "assert.h"
#define VERBOSE
int *mpMap;			/* the map of encoding and state */
#define MAX_NUM_STATE 8
struct character *tmpchar;	/*avoid mallocs */
int *transfer;
int *hamming;
int *occupied;
void
initMP (int num_genomes,
	struct genome_struct *genomes,
	struct genome_struct *labels, int num_genes)
{
  int i, j;

  for (i = 0; i < 2 * num_genomes + 6; i++)
    {
      if (i < num_genomes)
	{
	  genomes[i].chars = (struct character *)
	    malloc (sizeof (struct character) * (2 * num_genes + 2));
	  for (j = 0; j < 2 * num_genes + 2; j++)
	    {
	      initStates (&(genomes[i].chars[j]), MAX_NUM_STATE);
	    }
	}
      labels[i].chars = (struct character *)
	malloc (sizeof (struct character) * (2 * num_genes + 2));
      for (j = 0; j < 2 * num_genes + 2; j++)
	{
	  initStates (&(labels[i].chars[j]), MAX_NUM_STATE);
	}
    }
  for (i = 0; i < num_genomes; i++)
    {
      sequence2state (&genomes[i], num_genes);
      for (j = 0; j < num_genes; j++)
	printf ("%d", genomes[i].genes[j]);
      printf ("\n");
    }
  return;
}

void
Sequence2MP (struct genome_struct *genomes,
	     struct genome_struct *labels,
	     char *fname, int num_genomes, int num_genes)
{
  int i, j, g0;
  char *command;
  FILE *f1;

  /*create the map and clean it */
  mpMap = (int *) malloc (sizeof (int *) * MAX_NUM_STATE);
  for (j = 0; j < MAX_NUM_STATE; j++)
    mpMap[j] = -1;

  tmpchar = (struct character *) malloc (sizeof (struct character));
  initStates (tmpchar, MAX_NUM_STATE);

  transfer = (int *) malloc (sizeof (int) * (num_genes * 2 + 2));
  hamming = (int *) malloc (sizeof (int) * (num_genes * 2 + 2));
  occupied = (int *) malloc (sizeof (int) * (num_genes * 2 + 2));
  /*initializa all char state to zero */
  initMP (num_genomes, genomes, labels, num_genes);
}

void
cleanMP (struct genome_struct *genomes,
	 struct genome_struct *labels, int num_genomes, int num_genes)
{
  int i, j;
  for (i = 0; i < 2 * num_genomes + 6; i++)
    {
      if (i < num_genomes)
	{
	  for (j = 0; j < 2 * num_genes + 2; j++)
	    free (genomes[i].chars[j].states);
	  free (genomes[i].chars);
	}
      for (j = 0; j < 2 * num_genes + 2; j++)
	free (labels[i].chars[j].states);
      free (labels[i].chars);
    }

  free (mpMap);

  free (transfer);
  free (hamming);
  free (occupied);

  free (tmpchar->states);
  free (tmpchar);

}

void
clearInternal (struct tNode *tree, struct genome_struct *labels,
	       int num_genes, int num_genomes)
{
  /*add internal genomes from labels, and clean the state to all 0 */
  struct genome_struct *gen;
  int id, i;
  if (tree == NULL)
    return;
  if (tree->leaf == FALSE)
    {
      /* index into labels array using tree->tag;
       *            it's negative, so compute proper index */
      id = tree->tag;
      gen = tree->genome = &labels[num_genomes - id];
      gen->genome_num = id;
      for (i = 0; i < 2 * num_genes; i++)
	clearCharAll (&gen->chars[i]);
    }
  clearInternal (tree->lChild, labels, num_genes, num_genomes);
  clearInternal (tree->rChild, labels, num_genes, num_genomes);
  return;
}

void
initialize_tree_MPME (int COND, struct tNode *tree,
		      struct tNode *tpool,
		      struct genome_struct *labels, struct qNode *queue,
		      struct adj_struct *adj_list,
		      struct adj_struct *adj_pool, int **weights,
		      int *stack, int *degree, int *otherEnd,
		      intpair_t * neighbors, edge_t * edges,
		      int *incycle, int *outcycle, int *pred, int *succ,
		      int *code, int *decode, int num_genes,
		      int num_genomes, int inittspsolver, int thresh,
		      int CIRCULAR)
{
  /* c must be initialized before this was called */

  clearInternal (tree, labels, num_genes, num_genomes);
  /* clear the char string for all internal nodes */

  firstPass (tree->lChild, num_genes, num_genomes);

  secondPass (tree->lChild, num_genes, num_genomes);
}

/*should call from the child of root*/
void
firstPass (struct tNode *tree, int num_genes, int num_genomes)
{
  int i;
  if (tree == NULL || tree->leaf == TRUE)
    return;
  /*NULL, and leave should not do anything */

  if (tree->lChild != NULL)
    firstPass (tree->lChild, num_genes, num_genomes);
  if (tree->rChild != NULL)
    firstPass (tree->rChild, num_genes, num_genomes);

  if (tree->parent->parent == NULL)
    {
      /*child of the root node, do not need to do anything */
      if (tree->lChild->leaf == TRUE)
	{			/* lchild is a leaf */
	}
      else if (tree->rChild->leaf == TRUE)
	{			/*rChild is a leaf */
	}
      if (tree->rChild->leaf != TRUE && tree->lChild->leaf != TRUE)
	{
	  /*printf("Ooops, I think it should have at least a child of leaf\n"); */
	}
    }
  if (tree->parent == NULL)
    {
    }
  else
    {
      assert (tree->lChild != NULL && tree->rChild != NULL);

      /* for all characters, assign the states */
      for (i = 0; i < num_genes; i++)
	{
	  /*test for AND */
	  charAnd (&tree->genome->chars[i], &tree->lChild->genome->chars[i],
		   &tree->rChild->genome->chars[i]);

	  /*if AND is empty, use OR */
	  if (allZero (&tree->genome->chars[i]) == 1)
	    {
	      charOR (&tree->genome->chars[i],
		      &tree->lChild->genome->chars[i],
		      &tree->rChild->genome->chars[i]);
	    }
	}
    }
}

/*should call from the child of root*/
void
secondPass (struct tNode *tree, int num_genes, int num_genomes)
{
  int i, loc;
  int *octmp;
  if (tree == NULL || tree->leaf == TRUE)
    return;

  for (i = 0; i < 2 * num_genes + 2; i++)
    occupied[i] = 0;
  octmp = occupied + 1;
  octmp += num_genes;
  if (tree->parent == NULL)
    {				/*root node, do not need to do anything */
    }
  else
    {
      assert (tree->parent != NULL);

      /* for all characters, assign the states */
      for (i = 0; i < num_genes; i++)
	{
	  /*test for AND */
	  charAnd (tmpchar, &tree->genome->chars[i],
		   &tree->parent->genome->chars[i]);

	  /*if AND is empty, randompick */
	  if (allZero (tmpchar) == 1)
	    {
	      loc = randomPick (&tree->genome->chars[i]);
	      tree->genome->genes[i] = loc;
	      clearCharAll (&tree->genome->chars[i]);
	      setTrue (&tree->genome->chars[i], loc);
	    }
	  else
	    {
	      copyState (&tree->parent->genome->chars[i],
			 &tree->genome->chars[i]);
	      tree->genome->genes[i] = tree->parent->genome->genes[i];
	    }
	}
    }

  if (tree->lChild != NULL)
    secondPass (tree->lChild, num_genes, num_genomes);
  if (tree->rChild != NULL)
    secondPass (tree->rChild, num_genes, num_genomes);
}

int
compact (int state, int loc, int num_genomes)
{
  int i = 0;
  assert (i != num_genomes);
  return 0;
}

void
printMP (struct tNode *tree, int num_genes, int num_genomes)
{
  int i;
  if (tree == NULL)
    return;
  printf ("%d\n", tree->tag);
  for (i = 0; i < num_genes; i++)
    {
      printf ("%c", mpMap[tree->genome->genes[i]]);
    }
  printf ("\n");
  printMP (tree->lChild, num_genes, num_genomes);
  printMP (tree->rChild, num_genes, num_genomes);
}

void
sequence2state (struct genome_struct *genome, int num_genes)
{
  int i, j;
  for (i = 0; i < num_genes; i++)
    {
      for (j = 0; j < MAX_NUM_STATE; j++)
	{
	  if (mpMap[j] == genome->genes[i])
	    {
	      genome->genes[i] = j;
	      break;
	    }
	  if (mpMap[j] == -1)
	    {
	      mpMap[j] = genome->genes[i];
	      genome->genes[i] = j;
	      break;
	    }
	}
    }
  for (i = 0; i < num_genes; i++)
    {
      setTrue (&genome->chars[i], genome->genes[i]);
    }
  return;
}

void
setMPmatrix (int **distmatrix, struct genome_struct *genomes,
	     int num_genes, int num_genomes, distmem_t * distmem,
	     int CIRCULAR)
{
  int i, j;
  for (i = 0; i < num_genomes; i++)
    {
      distmatrix[i][i] = 0;
      for (j = i + 1; j < num_genomes; j++)
	{
	  distmatrix[i][j] = distmatrix[j][i] =
	    hammingMP (genomes + i, genomes + j, num_genes);
	}
    }
}

int
hammingMP (struct genome_struct *g1, struct genome_struct *g2, int num_genes)
{
  int i, score;
  score = 0;
  for (i = 0; i < num_genes; i++)
    {
      if (g1->genes[i] != g2->genes[i])
	score++;
    }
  return score;
}
